/*
avltree.c
12/17/04
andrew snowden
*/

#include <malloc.h>
#include <memory.h>
#include "avltree.h"

#define HEIGHT(node)  ((node)?((node)->height):0)
#define MAX(a, b)     (((a)>(b))?(a):(b))

typedef struct tag_node_t node_t;

struct tag_node_t {
  node_t *left, *right, *parent;
  node_t *prev, *next;
  void *data;
  int height;
};

struct tag_avl_t {
  node_t *root, *first, *last;
  int count;
  avl_cmp_t cmp;
  void *param;
};


/*** TREE OPERATIONS ***/


avl_t *avl_create(avl_cmp_t cmp, void *param) {
  avl_t *avl;

  avl=malloc(sizeof(avl_t));
  if (!avl) return NULL;
  memset(avl, 0, sizeof(avl_t));
  avl->cmp=cmp;
  avl->param=param;
  return avl;
}

void avl_destroy(avl_t *avl) {
  node_t *p, *tmp;

  for (p=avl->first; p; p=tmp) {
    tmp=p->next;
    free(p);
  }
  free(avl);
}

static node_t *new_node(void *data) {
  node_t *node;

  node=malloc(sizeof(node_t));
  if (!node) return NULL;
  memset(node, 0, sizeof(node_t));
  node->data=data;
  node->height=1;
  return node;
}

/* assumes node->right!=null */
static void rotate_left(avl_t *avl, node_t *node) {
  node_t *right=node->right, *tmp;

  tmp=right->left;
  right->left=node;
  node->right=tmp;
  if (tmp) tmp->parent=node;
  if (node->parent) {
    if (node->parent->left==node) node->parent->left=right;
    else node->parent->right=right;
  } else
    avl->root=right;
  right->parent=node->parent;
  node->parent=right;
  node->height=MAX(HEIGHT(node->left), HEIGHT(node->right))+1;
  right->height=MAX(node->height, HEIGHT(right->right))+1;
}

/* assumes node->left!=null */
static void rotate_right(avl_t *avl, node_t *node) {
  node_t *left=node->left, *tmp;

  tmp=left->right;
  left->right=node;
  node->left=tmp;
  if (tmp) tmp->parent=node;
  if (node->parent) {
    if (node->parent->left==node) node->parent->left=left;
    else node->parent->right=left;
  } else
    avl->root=left;
  left->parent=node->parent;
  node->parent=left;
  node->height=MAX(HEIGHT(node->left), HEIGHT(node->right))+1;
  left->height=MAX(HEIGHT(left->left), node->height)+1;
}

int avl_insert(avl_t *avl, void *data) {
  node_t *node, *cur, *tmp;
  int cmp, htl, htr;

/* if not root, make root */
  if (!avl->root) {
    avl->root=new_node(data);
    if (!avl->root) return 0;
    avl->first=avl->root;
    avl->last=avl->root;
    avl->count=1;
    return 1;
  }

/* search tree for spot to insert */
  cur=avl->root;
  while (cur) {
    tmp=cur;
    cmp=avl->cmp(avl->param, data, cur->data);
    if (cmp<0) cur=cur->left;
    else if (cmp>0) cur=cur->right;
    else return 0;
  }
  cur=tmp;

/* add new node */
  node=new_node(data);
  if (!node) return 0;
  if (cmp<0) {
    cur->left=node;
    node->parent=cur;
    node->prev=cur->prev;
    node->next=cur;
    if (cur->prev) cur->prev->next=node;
    else avl->first=node;
    cur->prev=node;
  } else {
    cur->right=node;
    node->parent=cur;
    node->prev=cur;
    node->next=cur->next;
    if (cur->next) cur->next->prev=node;
    else avl->last=node;
    cur->next=node;
  }

/* now rebalance */
  while (cur) {
    tmp=cur->parent;
    htl=HEIGHT(cur->left);
    htr=HEIGHT(cur->right);
    if (htl>htr+1) {
      if (HEIGHT(cur->left->right)>HEIGHT(cur->left->left))
        rotate_left(avl, cur->left);
      rotate_right(avl, cur);
    } else if (htr>htl+1) {
      if (HEIGHT(cur->right->left)>HEIGHT(cur->right->right))
        rotate_right(avl, cur->right);
      rotate_left(avl, cur);
    } else {
      cur->height=MAX(htl, htr)+1;
    }
    cur=tmp;
  }

/* done! */
  avl->count++;
  return 1;
}

int avl_count(avl_t *avl) {
  return avl->count;
}


/*** TREE ITERATION ***/


#define NODE(x) ((node_t *)(x))
#define VOID(x) ((void *)(x))

static int it_aux(avl_it_t *it) {
  it->end=(it->reserved==NULL);
  if (it->end) {
    it->data=NULL;
    return 0;
  } else {
    it->data=NODE(it->reserved)->data;
    return 1;
  }
}

void avl_it_init(avl_it_t *it, avl_t *avl) {
  memset(it, 0, sizeof(avl_it_t));
  it->avl=avl;
}

int avl_it_find(avl_it_t *it, void *data) {
  node_t *cur, *tmp;
  int cmp;

  cur=it->avl->root;
  if (!cur) {
    it->reserved=NULL;
    it_aux(it);
    return -1;
  }
  while (cur) {
    tmp=cur;
    cmp=it->avl->cmp(it->avl->param, cur->data, data);
    if (cmp<0)
      cur=cur->right;
    else if (cmp>0)
      cur=cur->left;
    else {
      it->reserved=(void *)cur;
      it_aux(it);
      return 0;
    }
  }
  cur=tmp;
  it->reserved=(void *)cur;
  it_aux(it);
  return cmp;
}

int avl_it_first(avl_it_t *it) {
  it->reserved=VOID(it->avl->first);
  return it_aux(it);
}

int avl_it_last(avl_it_t *it) {
  it->reserved=VOID(it->avl->last);
  return it_aux(it);
}

int avl_it_next(avl_it_t *it) {
  it->reserved=VOID(NODE(it->reserved)->next);
  return it_aux(it);
}

int avl_it_prev(avl_it_t *it) {
  it->reserved=VOID(NODE(it->reserved)->next);
  return it_aux(it);
}
