/*
rel8.c
12/30/04
andrew snowden

computes a basis for quadratic relations on eight points and shows that
these generated all relations of degree three and four.

this program calls matlab through the system function three times.  if
the command "matlab" does not run matlab then this program will fail.
*/

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <stdlib.h>
#include "avltree.h"
#include "tableaux.h"


/* global variables */
db_t *db;
tableau_t *tmp_deg_2_tab;
int quad_rel[14][105];
int cubic_rank, quartic_rank;



/**************************
* MISCELLANEOUS FUNCTIONS *
**************************/

/*
sorts a small array of integers.
*/
void sort_array(int *ar, int n) {
  int i, j, min, minpos, tmp;

  for (i=0; i<n; i++) {
    min=ar[i];
    minpos=i;
    for (j=i+1; j<n; j++) {
      if (ar[j]<min) {
        min=ar[j];
        minpos=j;
      }
    }
    tmp=ar[i];
    ar[i]=ar[minpos];
    ar[minpos]=tmp;
  }

}


/*
given a pair of integers 0 <= j <= i < 14 convert to a single integer.
the return value satisfies 0 <= ret < 105.  note that R_1 is 14
dimensional (i and j specify elements of its basis) and Sym^2(R_1) is
105 dimensional.
*/
int sym1_pair_to_sym2(int i, int j) {
  if (i<0 || i>=14 || j<0 || j>i) {
    printf("sym1_pair_to_sym2: bad range\n");
    exit(1);
  }
  return (i*(i+1))/2+j;
}


/*
inverse to above function.
*/
void sym2_to_sym1_pair(int n, int *i, int *j) {
  int k;

  if (n<0 || n>=105) {
    printf("sym2_to_sym1_pair: bad range\n");
    exit(1);
  }
  for (k=0; k*(k+1)<=2*n; k++);
  *i=k-1;
  *j=n-(k*(k-1))/2;
}


/*
given a triple i, j, k all >=0 and <14, but with no assumption on their
relative order, convert to a single integer.  the return value satisfies
0 <= ret < 660.  note that R_1 is 14 dimensional (i, j and k specify
elements of R_1) and Sym^3(R_1) is 560 dimensional.
*/
int sym1_triple_to_sym3(int i, int j, int k) {
  int ar[3];

  if (i<0 || i>=14 || j<0 || j>=14 || k<0 || k>=14) {
    printf("sym1_triple_to_sym3: bad range\n");
    exit(1);
  }
  ar[0]=i;
  ar[1]=j;
  ar[2]=k;
  sort_array(ar, 3);
  return (ar[2]*(ar[2]+1)*(ar[2]+2))/6 +
         (ar[1]*(ar[1]+1))/2 +
         ar[0];
}


/*
given a quartet i, j, k, l all >=0 and <14, but with no assumption on
their relative order, convert to a single integer.  the return value
satisfies 0 <= ret < 2380.  note that R_1 is 14 dimensional (i, j, k
and l specify elements of R_1) and Sym^4(R_1) is 2380 dimensional.
*/
int sym1_quartet_to_sym4(int i, int j, int k, int l) {
  int ar[4];

  if (i<0 || i>=14 || j<0 || j>=14 || k<0 || k>=14 || l<0 || l>=14) {
    printf("sym1_triple_to_sym3: bad range\n");
    exit(1);
  }
  ar[0]=i;
  ar[1]=j;
  ar[2]=k;
  ar[3]=l;
  sort_array(ar, 4);
  return (ar[3]*(ar[3]+1)*(ar[3]+2)*(ar[3]+3))/24 +
         (ar[2]*(ar[2]+1)*(ar[2]+2))/6 +
         (ar[1]*(ar[1]+1))/2 +
         ar[0];
}



/*
is string an integer
*/
int is_integer(char *str) {
  char *cp;

  if (*str=='\0') return 0;
  if (*str=='-') str++;
  for (cp=str; *cp; cp++) {
    if (!isdigit(*cp)) return 0;
  }
  return 1;
}


/*
removes spaces from beginning and end of string
*/
void trim_spaces(char *str) {
  int i, n;

  /* remove leading spaces */
  for (i=0; isspace(str[i]) && str[i]!=0; i++);
  if (str[i]=='\0') {
    str[0]='\0';
    return;
  }
  strcpy(str, str+i);

  /* remove tailing spaces */
  n=strlen(str);
  if (n==0) return;
  for (i=n-1; isspace(str[i]) && i>0; i--);
  if (i==0 && isspace(str[0])) {
    str[0]='\0';
    return;
  }
  str[i+1]='\0';

}

/*
return the product of two standard degree one tableau in terms of
the semi-standard basis.
*/
sum_t *compute_degree_one_product(int i, int j) {
  entry_t *ent1, *ent2;
  sum_t *sum;
  int k;

  /* build product tableau (just concatenate the two we have) */
  ent1=db->lookup[0][i];
  ent2=db->lookup[0][j];
  for (k=0; k<4; k++) {
    tmp_deg_2_tab->row[0][k]=ent1->tableau[0][k];
    tmp_deg_2_tab->row[1][k]=ent1->tableau[1][k];
    tmp_deg_2_tab->row[0][4+k]=ent2->tableau[0][k];
    tmp_deg_2_tab->row[1][4+k]=ent2->tableau[1][k];
  }

  /* make product standard */
  sum=tableau_standardize(db, tmp_deg_2_tab);
  if (!sum) {
    printf("\n\n ** error in standardizing **\n");
    exit(1);
  }

  /* done */
  return sum;

}



/************************
* 1. INITIALIZE PROGRAM *
************************/

/*
removes files that will be created, if they already exist, and
created the tableau database.
*/
void initialize_program() {

  /* say what we're doing */
  printf("initializing program...");
  fflush(stdout);

  /* remove files that this program will create */
  system("rm -f deg1_gen");
  system("rm -f quad_rel");

  /* create tableaux database */
  db=db_create(8);
  standard_add_all(db, 1, 1);
  standard_add_all(db, 2, 0);
  standard_add_all(db, 3, 0);
  standard_add_all(db, 4, 0);

  /* check that there are the right number of tableaux */
  if (db->stdcnt[0]!=14 || db->stdcnt[1]!=91 ||
      db->stdcnt[2]!=364 || db->stdcnt[3]!=1085) {
    printf("\n\n** wrong number of tableaux! **\n");
    exit(1);
  }

  /* allocate tmp_deg_2_tab */
  tmp_deg_2_tab=tableau_create_empty(8, 2);
  if (!tmp_deg_2_tab) {
    printf("\n\n** out of memory **\n");
    exit(1);
  }

  /* finish */
  printf("done\n");
  fflush(stdout);

}



/**********************************
* 2. OUTPUT DEGREE ONE GENERATORS *
**********************************/

/*
outputs the basis of R_1 that we will always use: the standard
tableau
*/
void output_degree_one_generators() {
  FILE *outf;
  int i, j;
  entry_t *ent;

  /* say what we're doing */
  printf("outputting degree one generators...");
  fflush(stdout);

  /* open output file */
  outf=fopen("deg1_gen", "wt");
  if (!outf) {
    printf("\n\n** could not open deg1_gen for output **\n");
    exit(1);
  }

  /* loop over degree one generators */
  for (i=0; i<14; i++) {

    /* output generator */
    ent=db->lookup[0][i];
    fprintf(outf, "[%d]\n", i+1);
    for (j=0; j<4; j++) {
      if (j!=0) fprintf(outf, " ");
      fprintf(outf, "%d", ent->tableau[0][j]);
    }
    fprintf(outf, "\n");
    for (j=0; j<4; j++) {
      if (j!=0) fprintf(outf, " ");
      fprintf(outf, "%d", ent->tableau[1][j]);
    }
    fprintf(outf, "\n\n");

  }

  /* finished */
  fclose(outf);
  printf("done\n");
  fflush(stdout);

}


/*****************************
* 3. LIST QUADRATIC PRODUCTS *
*****************************/

/*
loops over all pairs of standard weight one tableaux, computes their
product in terms of the semi-standard basis of weight two tableaux
and writes it to a matrix.

the resulting matrix represents the map Sym^2(R_1) -> R_2 given by
multiplication.  the columns are indexed by a basis for Sym^2(R_1)
and the rows by a basis for R_2.  thus the matrix has 105 columns
and 91 rows.  the kernel of this matrix is the space of quadratic
relations.
*/
void list_quadratic_products() {
  FILE *outf;
  int i, j, k;
  int row, col, value;
  entry_t *ent1, *ent2;
  sum_t *sum;

  /* say what we're doing */
  printf("computing matrix of Sym^2(R_1) -> R_2 ...");
  fflush(stdout);

  /* open output file */
  outf=fopen("matrix.dat", "wt");
  if (!outf) {
    printf("\n\n** could not open matrix.dat for output **\n");
    exit(1);
  }

  /* loop through all possible products */
  for (i=0; i<14; i++) {
    for (j=0; j<=i; j++) {

      /* compute product */
      sum=compute_degree_one_product(i, j);

      /* output row of matrix */
      col=sym1_pair_to_sym2(i, j)+1;
      for (k=0; k<sum->len; k++) {
        row=sum->id[k];
        value=sum->count[k];
        fprintf(outf, "%d %d %d\n", row, col, value);
      }
      sum_free(sum);

    }
  }

  /* finish up*/
  fclose(outf);
  printf("done\n");
  fflush(stdout);

}



/*********************************
* 4. COMPUTE QUADRATIC RELATIONS *
*********************************/

/*
runs the matlab script compute_kernel.m to compute the kernel of
the matrix contained in the file matrix.dat.
*/
void compute_quadratic_relations() {
  printf("computing quadratic relations in matlab...");
  fflush(stdout);
  system("rm -f matlab_out");
  system("matlab < compute_kernel.m > /dev/null");
  printf("done\n");
  fflush(stdout);
}



/************************
* 5. READ MATLAB OUTPUT *
************************/

/*
reads the output of convert_kenrel.m.  the output file is simply called
matlab_out.  it should contain 14 lines each of which contains a list of
105 integers.  the 14 lines are a basis for the 14 dimensional kernel of
the map Sym^2(R_1) -> R_2 while the columns specify the coordinates of
a kernel vector in the chosen basis for Sym^2(R_1).
*/
void read_matlab_output() {
  char lnbuf[1024];
  FILE *inf;
  char *cp, *cp2;
  int n, i, val, nonzero;

  /* say what we're doing */
  printf("reading output of matlab...");
  fflush(stdout);

  /* open matlab_out file */
  inf=fopen("matlab_out", "rt");
  if (!inf) {
    printf("\n\n** file 'matlab_out' not found **\n");
    exit(1);
  }

  /* read lines of file */
  n=0;
  while (fgets(lnbuf, sizeof(lnbuf), inf)) {

    /* if line is all whitespace, skip it */
    trim_spaces(lnbuf);
    if (*lnbuf=='\0') continue;

    /* if we've already read 14 lines at this point, there's a problem */
    if (n==14) {
      printf("\n\n** too many lines in output file **\n");
      exit(1);
    }

    /* parse comma list */
    cp=lnbuf;
    i=0;
    nonzero=0;
    while (cp) {
      if (i>=105) {
        printf("\n\n** output contains too many columns **\n");
        exit(1);
      }
      cp2=strchr(cp, ',');
      if (cp2) *cp2='\0';
      trim_spaces(cp);
      if (!is_integer(cp)) {
        printf("\n\n** output contains non-integral values **\n");
        exit(1);
      }
      val=atoi(cp);
      quad_rel[n][i]=val;
      if (val!=0) nonzero=1;
      if (cp2) cp=cp2+1;
      else cp=NULL;
      i++;
    }

    /* check that all is well */
    if (i!=105) {
      printf("\n\n** output contains too few columns **\n");
      exit(1);
    }
    if (!nonzero) {
      printf("\n\n** output contains zero row **\n");
      exit(1);
    }

    /* advance counter */
    n++;

  }

  /* finished */
  fclose(inf);
  printf("done\n");
  fflush(stdout);

}



/********************************
* 6. VERIFY QUADRATIC RELATIONS *
********************************/

/*
goes through each of the 14 quadratic relations that have been
computed and verify that they are actually valid relations.
*/
void verify_quadratic_relations() {
  sum_t *sum;
  int total[91];
  int n, i, a, b, v, k, vv, j;

  /* say what we're doing */
  printf("verifying quadratic relations...");
  fflush(stdout);

  /* loop through each relation */
  for (n=0; n<14; n++) {

    /* go through each term in the relation */
    memset(total, 0, sizeof(total));
    for (i=0; i<105; i++) {

      /* skip if coefficient is zero */
      v=quad_rel[n][i];
      if (v==0) continue;

      /* get image of element of sym2 in r2 in terms of standard basis */
      sym2_to_sym1_pair(i, &a, &b);
      sum=compute_degree_one_product(a, b);

      /* add this term to the total */
      for (k=0; k<sum->len; k++) {
        j=sum->id[k]-1;
        vv=sum->count[k];
        if (j<0 || j>=91) {
          printf("\n\n** degree two id out of range **\n");
          exit(1);
        }
        total[j]+=v*vv;
      }

    }

    /* we have just computed the image of the relation in r2; now
       make sure that it is zero */
    for (i=0; i<91; i++) {
      if (total[i]!=0) {
        printf("\n\n** relation is incorrect! **\n");
        exit(1);
      }
    }

  }

  /* finished */
  printf("done\n");

}



/********************************
* 7. OUTPUT QUADRATIC RELATIONS *
********************************/

/*
writes the array of quadratic relations to the file quad_rel
*/
void output_quadratic_relations() {
  FILE *outf;
  int n, i, a, b, v;
  int first;

  /* say what we're doing */
  printf("outputting quadratic relations...");
  fflush(stdout);

  /* open output file */
  outf=fopen("quad_rel", "wt");
  if (!outf) {
    printf("\n\n** could not open quad_rel for output **\n");
    exit(1);
  }

  /* loop over quadratic relations */
  for (n=0; n<14; n++) {

    /* go through each term in the sum */
    first=1;
    for (i=0; i<105; i++) {
      v=quad_rel[n][i];
      if (v==0) continue;
      sym2_to_sym1_pair(i, &a, &b);
      if (first) {
        first=0;
        fprintf(outf, "%d", v);
      } else if (v<0) {
        fprintf(outf, " - %d", -v);
      } else {
        fprintf(outf, " + %d", v);
      }
      fprintf(outf, "[%d][%d]", a, b);
    }
    fprintf(outf, "\n\n");

  }

  /* finished */
  fclose(outf);
  printf("done\n");
  fflush(stdout);

}



/************************************
* 8. LIST GENERATED CUBIC RELATIONS *
************************************/

/*
lists the cubic relations generated by the 14 quadratic relations.
this function creates a file caled cubic.dat which is a sparse matrix.
the rows of the matrix index a basis of Sym^3(R_1) (this space is 560
dimensional).  the columns correspond to the cubic relations generated
by the quadratics.  the rank of this matrix is the dimension of the space
of cubic relations which are generated by quadratic relations.
*/
void list_generated_cubic_relations() {
  FILE *outf;
  int total[560];
  int n, i, k, a, b, v, r, col=1;

  /* say what we're doing */
  printf("outputting generated cubic relations...");
  fflush(stdout);

  /* open output file */
  outf=fopen("matrix.dat", "wt");
  if (!outf) {
    printf("\n\n** could not open matrix.dat for output **\n");
    exit(1);
  }

  /* loop over quadratic relations */
  for (n=0; n<14; n++) {

    /* loop over degree one elements */
    for (i=0; i<14; i++) {

      /* compute product of relation n with element of sym1 given by i */
      memset(total, 0, sizeof(total));
      for (k=0; k<105; k++) {

        /* skip term if coefficient is zero */
        v=quad_rel[n][k];
        if (v==0) continue;

        /* get element of sym3 */
        sym2_to_sym1_pair(k, &a, &b);
        r=sym1_triple_to_sym3(a, b, i);

        /* add to total */
        total[r]+=v;

      }

      /* write element of sym3 to file */
      for (k=0; k<560; k++) {
        if (total[k]==0) continue;
        fprintf(outf, "%d %d %d\n", k+1, col, total[k]);
      }

      /* move to next column */
      col++;

    }
  }

  /* finished */
  fclose(outf);
  printf("done\n");

}



/**************************************
* 9. LIST GENERATED QUARTIC RELATIONS *
**************************************/

/*
like previous function but for quartics.  note that Sym^4(R_1) has
dimension 2380.
*/
void list_generated_quartic_relations() {
  FILE *outf;
  int total[2380];
  int n, i, j, k, a, b, v, r, col=1;

  /* say what we're doing */
  printf("outputting generated quartic relations...");
  fflush(stdout);

  /* open output file */
  outf=fopen("matrix.dat", "wt");
  if (!outf) {
    printf("\n\n** could not open matrix.dat for output **\n");
    exit(1);
  }

  /* loop over quadratic relations */
  for (n=0; n<14; n++) {

    /* loop over sym2 */
    for (i=0; i<14; i++) {
      for (j=0; j<=i; j++) {

        /* compute product of relation n with element of sym1 given by i */
        memset(total, 0, sizeof(total));
        for (k=0; k<105; k++) {

          /* skip term if coefficient is zero */
          v=quad_rel[n][k];
          if (v==0) continue;

          /* get element of sym3 */
          sym2_to_sym1_pair(k, &a, &b);
          r=sym1_quartet_to_sym4(a, b, i, j);

          /* add to total */
          total[r]+=v;

        }

        /* write element of sym3 to file */
        for (k=0; k<2380; k++) {
          if (total[k]==0) continue;
          fprintf(outf, "%d %d %d\n", k+1, col, total[k]);
        }

        /* move to next column */
       col++;

      }
    }
  }

  /* finished */
  fclose(outf);
  printf("done\n");

}



/*****************************
* 10. COMPUTE RANK OF MATRIX *
*****************************/

/*
computes the rank of the matrix in matrix.dat by running the matlab
script compute_rank.m.
*/
int compute_rank() {
  FILE *inf;
  char lnbuf[128];
  int ret;

  /* say what we're doing */
  printf("computing rank in matlab...");
  fflush(stdout);

  /* run matlab */
  system("rm -f matlab_out");
  system("matlab < compute_rank.m > /dev/null");

  /* read result */
  inf=fopen("matlab_out", "rt");
  if (!inf) {
    printf("\n\n** file 'matlab_out' not found **\n");
    exit(1);
  }
  lnbuf[0]='\0';
  fgets(lnbuf, sizeof(lnbuf), inf);
  fclose(inf);
  trim_spaces(lnbuf);
  if (!is_integer(lnbuf)) {
    printf("\n\n** output is not an integer **\n");
    exit(1);
  }
  ret=atoi(lnbuf);
  
  /* finished */
  printf("done\n");
  fflush(stdout);
  return ret;

}



/*******************
* 11. SHOW RESULTS *
*******************/

void show_results() {

  printf("\nsummary of results:\n");

  printf("\n");
  printf("R_1 has dimension 14; a basis has been written to deg1_gen\n");
  printf("I_2 has dimension 14; a basis has been written to quad_rel\n");

  printf("\n");
  printf("Sym^3(R_1) has dimension 560; R_3 has dimension 364\n");
  printf("thus I_3 has dimension 196\n");
  printf("the subspace of I_3 generated by I_2 has dimension %d\n", cubic_rank);
  if (cubic_rank==196) printf("therefore I_2 generates I_3\n");
  else printf("therefore I_2 does NOT generate I_3\n");

  printf("\n");
  printf("Sym^4(R_1) has dimension 2380; R_4 has dimension 1085\n");
  printf("thus I_4 has dimension 1295\n");
  printf("the subspace of I_4 generated by I_2 has dimension %d\n", quartic_rank);
  if (quartic_rank==1295) printf("therefore I_2 generates I_4\n");
  else printf("therefore I_2 does NOT generate I_4\n");

  printf("\n");

}



/*****************
* MAIN PROCEDURE *
******************/

int main() {

  printf("\n");
  initialize_program();
  output_degree_one_generators();
  list_quadratic_products();
  compute_quadratic_relations();
  read_matlab_output();
  verify_quadratic_relations();
  output_quadratic_relations();
  list_generated_cubic_relations();
  cubic_rank=compute_rank();
  list_generated_quartic_relations();
  quartic_rank=compute_rank();
  show_results();
  return 0;

}

