/*
tab-db.c
andrew snowden

history:
12/30/04 - tableaux.c taken from older tableaux.c file
01/23/07 - split off tab-db.c
*/

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <stdlib.h>
#include "avltree.h"
#include "tableaux.h"


static int entry_cmp(void *param, entry_t *x1, entry_t *x2) {
  int sides=(int)param;
  int i, width;

  if (x1->wt<x2->wt) return -1;
  if (x1->wt>x2->wt) return 1;
  width=x1->wt*(sides/2);
  for (i=0; i<width; i++) {
    if (x1->tableau[0][i]<x2->tableau[0][i]) return -1;
    if (x1->tableau[0][i]>x2->tableau[0][i]) return 1;
  }
  for (i=0; i<width; i++) {
    if (x1->tableau[1][i]<x2->tableau[1][i]) return -1;
    if (x1->tableau[1][i]>x2->tableau[1][i]) return 1;
  }
  return 0;
}

entry_t *db_find_by_tableau(db_t *db, tableau_t *tableau) {
  entry_t tmp;
  avl_it_t it;

  if (db->sides!=tableau->sides) return NULL;
  if (tableau->weight<1 || tableau->weight>MAXWEIGHT) return NULL;
  if (db->avl[tableau->weight-1]==NULL) return NULL;
  memset(&tmp, 0, sizeof(entry_t));
  tmp.tableau[0]=tableau->row[0];
  tmp.tableau[1]=tableau->row[1];
  tmp.wt=tableau->weight;
  avl_it_init(&it, db->avl[tmp.wt-1]);
  if (avl_it_find(&it, (void *)&tmp)!=0) return NULL;
  return (entry_t *)avl_it_data(&it);
}

entry_t *db_lookup_std(db_t *db, int weight, int id) {
  int i;

  if (weight<1 || weight>MAXWEIGHT) return NULL;
  i=weight-1;
  if (db->lookup[i]==NULL) return NULL;
  if (id<1 || id>db->stdcnt[i]) return NULL;
  return db->lookup[i][id-1];
}

int db_add(db_t *db, tableau_t *tableau) {
  entry_t *x;
  int width, i;

  if (tableau->sides!=db->sides) return 0;
  if (tableau->weight<1 || tableau->weight>MAXWEIGHT) return 0;

  i=tableau->weight-1;
  if (!db->avl[i]) {
    db->avl[i]=avl_create((avl_cmp_t)entry_cmp, (void *)db->sides);
    if (!db->avl[i]) return 0;
    db->maxid[i]=1;
  }

  width=tableau->width;
  x=(entry_t *)malloc(sizeof(entry_t)+2*width);
  if (!x) return 0;
  x->tableau[0]=(unsigned char *)(x+1);
  x->tableau[1]=x->tableau[0]+width;
  x->wt=tableau->weight;
  x->id=db->maxid[i]++;
  memcpy(x->tableau[0], tableau->row[0], width);
  memcpy(x->tableau[1], tableau->row[1], width);

  return avl_insert(db->avl[i], x);
}

static void db_free_tree(avl_t *avl) {
  avl_it_t it;
  entry_t *p;

  avl_it_init(&it, avl);
  avl_it_first(&it);
  while (!avl_it_end(&it)) {
    p=(entry_t *)avl_it_data(&it);
    free(p);
    avl_it_next(&it);
  }
  avl_destroy(avl);
}

void db_free(db_t *db) {
  int i;

  for (i=0; i<MAXWEIGHT; i++) {
    if (db->avl[i]) db_free_tree(db->avl[i]);
    if (db->lookup[i]) free(db->lookup[i]);
  }
  free(db);
}

db_t *db_create(int sides) {
  db_t *db;

  db=(db_t *)malloc(sizeof(db_t));
  memset(db, 0, sizeof(db_t));
  db->sides=sides;
  return db;
}

void db_print_entry(int sides, entry_t *p, FILE *outf) {
  int i, width;

  fprintf(outf, "[%d]\n", p->id);
  width=sides*p->wt/2;
  for (i=0; i<width; i++) {
    fprintf(outf, "%d", p->tableau[0][i]);
    if (i<width-1) fputc(',', outf);
  }
  fputc('\n', outf);
  for (i=0; i<width; i++) {
    fprintf(outf, "%d", p->tableau[1][i]);
    if (i<width-1) fputc(',', outf);
  }
  fputc('\n', outf);
}

void db_write_entry(int sides, entry_t *p, FILE *outf) {
  int i, width;

  fprintf(outf, "%d, %d; ", sides, (int)p->wt);
  width=sides*p->wt/2;
  for (i=0; i<width; i++) {
    fprintf(outf, "%d", p->tableau[0][i]);
    if (i<width-1) fprintf(outf, ", ");
    else fprintf(outf, "; ");
  }
  for (i=0; i<width; i++) {
    fprintf(outf, "%d", p->tableau[1][i]);
    if (i<width-1) fprintf(outf, ", ");
    else fprintf(outf, ";\n");
  }
}

void db_print_all(db_t *db, int weight, FILE *outf) {
  avl_it_t it;
  entry_t *p;

  if (weight<1 || weight>MAXWEIGHT) return;
  if (db->avl[weight-1]==NULL) return;
  avl_it_init(&it, db->avl[weight-1]);
  avl_it_first(&it);
  while (!avl_it_end(&it)) {
    p=(entry_t *)avl_it_data(&it);
    db_print_entry(db->sides, p, outf);
    putc('\n', outf);
    avl_it_next(&it);
  }
}

void db_write_all(db_t *db, int weight, FILE *outf) {
  avl_it_t it;
  entry_t *p;

  if (weight<1 || weight>MAXWEIGHT) return;
  if (db->avl[weight-1]==NULL) return;
  avl_it_init(&it, db->avl[weight-1]);
  avl_it_first(&it);
  while (!avl_it_end(&it)) {
    p=(entry_t *)avl_it_data(&it);
    db_write_entry(db->sides, p, outf);
    avl_it_next(&it);
  }
}

entry_t *db_find_by_id(db_t *db, int weight, int id) {
  avl_it_t it;
  entry_t *p;

  if (weight<1 || weight>MAXWEIGHT) return NULL;
  if (db->avl[weight-1]==NULL) return NULL;
  avl_it_init(&it, db->avl[weight-1]);
  avl_it_first(&it);
  while (!avl_it_end(&it)) {
    p=(entry_t *)avl_it_data(&it);
    if (p->id==id) return p;
    avl_it_next(&it);
  }
  return NULL;
}

